@@ -1,11 +1,12 @@
 from pathlib import Path
 
 import numpy as np
-import xobjects as xo
 from scipy.special import factorial
 
+import xobjects as xo
+import xpart as xp
+
 from ..base_element import BeamElement
-from ..particles import ParticlesData
 from ..general import _pkg_root
 
 class ReferenceEnergyIncrease(BeamElement):
@@ -18,6 +19,10 @@ class ReferenceEnergyIncrease(BeamElement):
     _xofields = {
         'Delta_p0c': xo.Float64}
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(Delta_p0c=-self.Delta_p0c,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 ReferenceEnergyIncrease.XoStruct.extra_sources = [
         _pkg_root.joinpath('beam_elements/elements_src/referenceenergyincrease.h')]
 
@@ -30,6 +35,11 @@ class Drift(BeamElement):
 
     _xofields = {
         'length': xo.Float64}
+    isthick=True
+
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(length=-self.length,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
 
 Drift.XoStruct.extra_sources = [
         _pkg_root.joinpath('beam_elements/elements_src/drift.h')]
@@ -48,6 +58,13 @@ class Cavity(BeamElement):
         'lag': xo.Float64,
         }
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              voltage=-self.voltage,
+                              frequency=self.frequency,
+                              lag=self.lag,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 Cavity.XoStruct.extra_sources = [
         _pkg_root.joinpath('headers/constants.h'),
         _pkg_root.joinpath('beam_elements/elements_src/cavity.h')]
@@ -65,6 +82,11 @@ class XYShift(BeamElement):
         'dy': xo.Float64,
         }
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              dx=-self.dx, dy=-self.dy,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 XYShift.XoStruct.extra_sources = [
         _pkg_root.joinpath('beam_elements/elements_src/xyshift.h')]
 
@@ -94,6 +116,15 @@ class Elens(BeamElement):
         self.elens_length    = elens_length
         self.voltage         = voltage
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              current=self.current,
+                              inner_radius=self.inner_radius,
+                              outer_radius=self.outer_radius,
+                              elens_length=-self.elens_length,
+                              voltage=self.voltage,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 Elens.XoStruct.extra_sources = [
     _pkg_root.joinpath('beam_elements/elements_src/elens.h')]
 
@@ -112,6 +143,11 @@ class SRotation(BeamElement):
         'sin_z': xo.Float64,
         }
 
+    def to_dict(self):
+        dct = super().to_dict()
+        dct['angle'] = self.angle
+        return dct
+
     def __init__(self, angle=0, **nargs):
         anglerad = angle / 180 * np.pi
         nargs['cos_z']=np.cos(anglerad)
@@ -122,6 +158,11 @@ class SRotation(BeamElement):
     def angle(self):
         return np.arctan2(self.sin_z, self.cos_z) * (180.0 / np.pi)
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              angle=-self.angle,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 SRotation.XoStruct.extra_sources = [
         _pkg_root.joinpath('beam_elements/elements_src/srotation.h')]
 
@@ -146,6 +187,13 @@ class Multipole(BeamElement):
         'bal': xo.Float64[:],
         }
 
+    def to_dict(self):
+        dct = super().to_dict()
+        dct['knl'] = self.knl
+        dct['ksl'] = self.ksl
+        return dct
+
+
     def __init__(self, order=None, knl=None, ksl=None, bal=None, **kwargs):
 
         if bal is None and (
@@ -190,28 +238,49 @@ class Multipole(BeamElement):
             kwargs["bal"] = bal
             kwargs["order"] = (len(bal) - 2) // 2
 
-
-        # TODO: Remove when xobjects is fixed
-        kwargs["bal"] = list(kwargs['bal'])
-        self._temp_bal_length = len(kwargs['bal'])
+        ## TODO: Remove when xobjects is fixed
+        if 'bal' in kwargs.keys():
+            keep_bal = kwargs['bal']
+            kwargs['bal'] = len(keep_bal) # Initializes with zeros
 
         self.xoinitialize(**kwargs)
 
+        ## TODO: Remove when xobjects is fixed
+        if 'bal' in kwargs.keys():
+            if isinstance(keep_bal, np.ndarray):
+                keep_bal = self._buffer.context.nparray_to_context_array(
+                                                                    keep_bal)
+            self.bal[:]= keep_bal # Sets the values
+
     @property
     def knl(self):
-        idxes = np.array([ii for ii in range(0, self._temp_bal_length, 2)])
+        bal_length = len(self.bal)
+        idxes = np.array([ii for ii in range(0, bal_length, 2)])
         return [self.bal[idx] * factorial(idx // 2, exact=True) for idx in idxes]
 
     @property
     def ksl(self):
-        idxes = np.array([ii for ii in range(0, self._temp_bal_length, 2)])
+        bal_length = len(self.bal)
+        idxes = np.array([ii for ii in range(0, bal_length, 2)])
         return [self.bal[idx + 1] * factorial(idx // 2, exact=True) for idx in idxes]
         #idx = np.array([ii for ii in range(0, len(self.bal), 2)])
         #return self.bal[idx + 1] * factorial(idx // 2, exact=True)
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              order=self.order,
+                              length=-self.length,
+                              hxl=-self.hxl,
+                              hyl=-self.hyl,
+                              radiation_flag=0, #TODO, I force radiation off for now
+                              bal=-self.bal, # TODO: maybe it can be made more efficient
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 Multipole.XoStruct.extra_sources = [
-    _pkg_root.joinpath('random_number_generator/rng_src/base_rng.h'),
-    _pkg_root.joinpath('random_number_generator/rng_src/local_particle_rng.h'),
+    xp.general._pkg_root.joinpath(
+        'random_number_generator/rng_src/base_rng.h'),
+    xp.general._pkg_root.joinpath(
+        'random_number_generator/rng_src/local_particle_rng.h'),
     _pkg_root.joinpath('beam_elements/elements_src/multipole.h')]
 
 
@@ -244,6 +313,14 @@ class RFMultipole(BeamElement):
         'phase': xo.Float64[:],
     }
 
+    def to_dict(self):
+        dct = super().to_dict()
+        dct['knl'] = self.knl
+        dct['ksl'] = self.ksl
+        dct['pn'] = self.pn
+        dct['ps'] = self.ps
+        return dct
+
     def __init__(
         self,
         order=None,
@@ -252,9 +329,12 @@ class RFMultipole(BeamElement):
         pn=None,
         ps=None,
         bal=None,
-        p=None,
+        phase=None,
         **kwargs
     ):
+
+        assert 'p' not in kwargs, "`p` in RF Multipole is not supported anymore"
+
         if bal is None and (
             knl is not None
             or ksl is not None
@@ -306,46 +386,53 @@ class RFMultipole(BeamElement):
 
             order = n - 1
             bal = np.zeros(2 * order + 2)
-            p = np.zeros(2 * order + 2)
+            phase = np.zeros(2 * order + 2)
 
             idx = np.array([ii for ii in range(0, len(knl))])
             inv_factorial = 1.0 / factorial(idx, exact=True)
             bal[0::2] = knl * inv_factorial
             bal[1::2] = ksl * inv_factorial
 
-            p[0::2] = pn
-            p[1::2] = ps
+            phase[0::2] = pn
+            phase[1::2] = ps
 
             kwargs["bal"] = bal
-            kwargs["phase"] = p
+            kwargs["phase"] = phase
             kwargs["order"] = order
 
         elif (
             bal is not None
             and bal
-            and len(bal) > 2
+            and len(bal) >= 2
             and ((len(bal) % 2) == 0)
-            and p is not None
-            and p
-            and len(p) > 2
-            and ((len(p) % 2) == 0)
+            and phase is not None
+            and phase
+            and len(phase) >= 2
+            and ((len(phase) % 2) == 0)
         ):
             kwargs["bal"] = bal
-            kwargs["phase"] = p
+            kwargs["phase"] = phase
             kwargs["order"] = (len(bal) - 2) / 2
+        elif '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
+            pass
+        else:
+            raise ValueError('RF Multipole Invalid input!')
 
 
-        temp_bal = kwargs["bal"]
-        temp_phase = kwargs["phase"]
+        if '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
+            super().__init__(**kwargs)
+        else:
+            temp_bal = kwargs["bal"]
+            temp_phase = kwargs["phase"]
 
-        kwargs["bal"] = len(temp_bal)
-        kwargs["phase"] = len(temp_phase)
+            kwargs["bal"] = len(temp_bal)
+            kwargs["phase"] = len(temp_phase)
 
-        super().__init__(**kwargs)
+            super().__init__(**kwargs)
 
-        ctx = self._buffer.context
-        self.bal[:] = ctx.nparray_to_context_array(temp_bal)
-        self.phase[:] = ctx.nparray_to_context_array(temp_phase)
+            ctx = self._buffer.context
+            self.bal[:] = ctx.nparray_to_context_array(temp_bal)
+            self.phase[:] = ctx.nparray_to_context_array(temp_phase)
 
     @property
     def knl(self):
@@ -367,12 +454,12 @@ class RFMultipole(BeamElement):
 
     @property
     def pn(self):
-        idx = np.array([ii for ii in range(0, len(self.p), 2)])
+        idx = np.array([ii for ii in range(0, len(self.phase), 2)])
         return self.phase[idx]
 
     @property
     def ps(self):
-        idx = np.array([ii for ii in range(0, len(self.p), 2)])
+        idx = np.array([ii for ii in range(0, len(self.phase), 2)])
         return self.phase[idx + 1]
 
     def set_pn(self, value, order):
@@ -383,6 +470,16 @@ class RFMultipole(BeamElement):
         assert order <= self.order
         self.phase[order * 2 + 1] = value
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              order=self.order,
+                              voltage=-self.voltage,
+                              frequency=self.frequency,
+                              lag=self.lag,
+                              bal=[-bb for bb in self.bal], # TODO: maybe it can be made more efficient
+                              p = [pp for pp in self.phase],
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 RFMultipole.XoStruct.extra_sources = [
         _pkg_root.joinpath('headers/constants.h'),
         _pkg_root.joinpath('beam_elements/elements_src/rfmultipole.h')]
@@ -403,6 +500,14 @@ class DipoleEdge(BeamElement):
             'r43': xo.Float64,
             }
 
+    def to_dict(self):
+        dct = super().to_dict()
+        dct['h'] = self.h
+        dct['e1'] = self.e1
+        dct['hgap'] = self.hgap
+        dct['fint'] = self.fint
+        return dct
+
     def __init__(
         self,
         r21=None,
@@ -437,6 +542,11 @@ class DipoleEdge(BeamElement):
             assert not np.isclose(np.absolute(np.cos(e1 - temp)), ZERO)
             r43 = -h * np.tan(e1 - temp)
 
+            self.h = h
+            self.e1 = e1
+            self.hgap = hgap
+            self.fint = fint
+
         if r21 is not None and r43 is not None:
             kwargs['r21'] = r21
             kwargs['r43'] = r43
@@ -447,6 +557,12 @@ class DipoleEdge(BeamElement):
                 " or suitable values for h, e1, hgap, and fint provided"
             )
 
+    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        return self.__class__(
+                              r21=-self.r21,
+                              r43=-self.r43,
+                              _context=_context, _buffer=_buffer, _offset=_offset)
+
 
 DipoleEdge.XoStruct.extra_sources = [
         _pkg_root.joinpath('beam_elements/elements_src/dipoleedge.h')]
